@@ -1,10 +1,16 @@
+import warnings
+warnings.simplefilter(action='ignore', category=FutureWarning)
 import pandas as pd
 from datetime import datetime
 from sqlalchemy import create_engine, types, MetaData, Table, select, delete, and_, Column, PrimaryKeyConstraint
 from sqlalchemy.exc import NoSuchTableError, NoSuchColumnError
 from sqlalchemy import schema as sqlalchemy_schema
+from sqlalchemy import text as sql_text
+from sqlalchemy import inspect
 import numpy as np
 import warnings
+from tqdm import tqdm
+
 warnings.simplefilter(action='ignore', category=UserWarning)
 
 class Table_analyzer():
@@ -36,7 +42,7 @@ class Table_analyzer():
                 if date_has_time(column_data):
                     return "DATETIME"
                 return "DATE"
-            except ValueError:
+            except :
                 return False
             
         column_data = df[column_name]
@@ -87,7 +93,7 @@ class Table_analyzer():
         non_null_values = column_data.dropna()
 
         # Calculate the maximum length of the texts in the list
-        max_text_length = max(len(text.encode()) for text in non_null_values)
+        max_text_length = max(len(str(text).encode()) for text in non_null_values)
 
         # Calculate the suitable n value considering the buffer and the max_length constraint
         if 0 <= buffer < 1:
@@ -136,12 +142,22 @@ class Table_analyzer():
 
 
 class PySQL():
-    def __init__(self):
+    def __init__(self, if_error='pass'):
+        """
+        Init
+
+        Args:
+            if_error(str): pass / raise default=pass
+            
+        Returns:
+            PySQL object.
+        """
         self.log_dtypes =  {'function':types.VARCHAR(50), 'state':types.VARCHAR(50), 'log':types.VARCHAR(2000), 'connection_user':types.VARCHAR(50), 'process_id':types.INT(), 'datetime':types.DATETIME()}
-        self.dtypes_types = {'table':types.VARCHAR(50), 'schema':types.VARCHAR(50), 'dtypes_str':types.VARCHAR(4000), 'process_id':types.INT()}
+        self.dtypes_types = {'table':types.VARCHAR(50), 'schema':types.VARCHAR(50), 'dtypes_str':types.NVARCHAR(4000), 'process_id':types.INT()}
         self.process_id = -1
         self.dtypes = {}
-        
+        self.if_error = if_error
+    
     def _log_decorator(func):  
         def wrapper(self, *args, **kwargs):
             try:
@@ -153,6 +169,8 @@ class PySQL():
             except Exception as Error:
                 print(Error)
                 self.logger(func.__name__ , 'progress', str(Error))
+                if self.if_error == 'raise':
+                    raise Error
         return wrapper    
     
     def logger(self, func, state, log):
@@ -171,15 +189,17 @@ class PySQL():
         """
         if len(log)> 2000:
             log = log[:1999]
-        if 'config' not in self.engine.dialect.get_schema_names(self.engine):  # check for config schema
-            self.engine.execute(sqlalchemy_schema.CreateSchema('config'))
+        if 'config' not in self.inspector.get_schema_names():  # check for config schema
+            with self.engine.connect() as conn:
+                conn.execute(sqlalchemy_schema.CreateSchema('config'))
+                conn.commit()
             print('config schema not exist! \n new created!')
         now = datetime.now().isoformat().replace('T', ' ').split('.')[0]
         log_data = {'function':func, 'state':state, 'log':log,'connection_user':self.username ,'process_id':self.process_id, 'datetime':now}
         self.log_data = pd.DataFrame([log_data])
         self.log_data.to_sql('log', con=self.engine, schema='config', if_exists='append', dtype=self.log_dtypes, index=False)
         
-    def create_connection(self, server, database, username, password, port=1433):
+    def create_connection(self, server=None, database='temp', username=None, password=None, port=1433, local_sql=False):
         """
         creates a pysql.engine connection to your target database
         in order to have best experience to use, your sql_user(username) must to have datawriter, datareader and ddladmin 
@@ -189,24 +209,55 @@ class PySQL():
             username (str): username 
             password (str): password 
             port (str): connection port default : 1433
-            
+            local_sql (bool): if True PySQL try to connect to local sql_server with windows auth
         Returns:
             None
         """
-        self.server = server
-        self.database = database
-        self.username = username
-        self.password = password
-        self.port = port
-        self.connection_str = f"mssql+pyodbc://{self.username}:{self.password}@{self.server}:{self.port}/{self.database}?driver=ODBC+Driver+17+for+SQL+Server"
-        self.engine = create_engine(self.connection_str)
-        self.logger('create_connection', 'success', 'connected')
-        self.lastlog = self.read_sql_table('log', schema='config', columns=['process_id'])
-        self.process_id = self.lastlog.process_id.values.tolist()
-        self.process_id = max([int(i) for i in self.process_id])
+        if local_sql:
+            self.database = database
+            self.username = 'WINDOWS_AUTH'
+            self.connection_str = f'mssql+pyodbc://localhost/{self.database}?driver=ODBC+Driver+17+for+SQL+Server'
+            self.engine = create_engine(self.connection_str)
+            self.inspector = inspect(self.engine)
+            self.logger('create_connection', 'success', 'connected')
+            self.lastlog = self.read_sql_query("SELECT [process_id]  FROM [config].[log]")
+            self.process_id = self.lastlog.process_id.values.tolist()
+            self.process_id = max([int(i) for i in self.process_id])
+        else:
+            self.server = server
+            self.database = database
+            self.username = username
+            self.password = password
+            self.port = port
+            self.connection_str = f"mssql+pyodbc://{self.username}:{self.password}@{self.server}:{self.port}/{self.database}?driver=ODBC+Driver+17+for+SQL+Server"
+            self.engine = create_engine(self.connection_str)
+            self.inspector = inspect(self.engine)
+            self.logger('create_connection', 'success', 'connected')
+            self.lastlog = self.read_sql_query("SELECT [process_id]  FROM [config].[log]")
+            self.process_id = self.lastlog.process_id.values.tolist()
+            self.process_id = max([int(i) for i in self.process_id])
+
+    def n_batches_dataframe(self, dataframe, num_batches):
+        # Calculate the total number of rows in the DataFrame
+        total_rows = len(dataframe)
+        
+        # Calculate the batch size based on the number of batches
+        batch_size = total_rows // num_batches
+        
+        # Create a tqdm instance to display the progress bar
+        pbar = tqdm(total=num_batches, desc="sending to SQL")
+        
+        # Iterate through the DataFrame and yield the specified number of batches
+        for i in range(num_batches):
+            start_idx = i * batch_size
+            end_idx = start_idx + batch_size if i < num_batches - 1 else total_rows
+            yield dataframe.iloc[start_idx:end_idx]
+            pbar.update(1)  # Update the progress bar
+        pbar.close()  # Close the progress bar when done
+        
 
     @_log_decorator 
-    def to_sql(self, df, table_name, schema=None, if_exists='append', index=True, index_label=None, primary_key=None, chunksize=1, date_normalizer=True, text_cutter=True):
+    def to_sql(self, df, table_name, schema=None, if_exists='append', index=True, index_label=None, method=None, primary_key=None, chunksize=1, date_normalizer=True, text_cutter=True, verbose=False):
         """
         Write records stored in a DataFrame to a SQL database.
     
@@ -228,6 +279,8 @@ class PySQL():
             Column label for index column(s). If None is given (default) and
             `index` is True, then the index names are used.
             A sequence should be given if the DataFrame uses MultiIndex.
+        method : str , default None
+            'multi' for insert multi row in one query
         primary_key : str , optional
             it uses pysql.set_primary_key so :
             just one time allowed (if PK exists error!)
@@ -237,7 +290,8 @@ class PySQL():
             By default, all rows will be written at once.
         date_normalizer : bool, default True
             its run PySQL.date_normalizer and make date formats storable foe sql server
-    
+        verbose :  bool, default False
+            prgress bar for data sending
     
             .. versionadded:: 1.3.0
     
@@ -250,18 +304,41 @@ class PySQL():
             Number of rows affected by to_sql. None is returned if the callable
             passed into ``method`` does not return the number of rows.
         """
+        row_number = 0
         if schema!= None:
-            if schema not in self.engine.dialect.get_schema_names(self.engine):  # check for schema existance
-                self.engine.execute(sqlalchemy_schema.CreateSchema(schema))
+            if schema not in self.inspector.get_schema_names():  # check for schema existance
+                with self.engine.connect() as conn:
+                    conn.execute(sqlalchemy_schema.CreateSchema(schema))
+                    conn.commit()
         df['process_id'] = [self.process_id]*len(df)
         self.dtypes = self.dtypes | {'process_id':types.INT()}
         if date_normalizer:
             df = self.date_normalizer(df)
+
         if text_cutter:
             df = self.text_cutter(df)
-        
-        self.df = df
-        row_number = df.to_sql(name=table_name, con=self.engine, schema=schema, if_exists=if_exists, index=index, index_label=index_label, dtype=self.dtypes, chunksize=chunksize)
+
+        if verbose:
+            if_exists_flag = False
+            self.df = df
+            df_length = len(df)
+
+            if df_length > 1000:
+                n_batches = 1000
+                batches = self.n_batches_dataframe(df, n_batches)
+                for batch in batches:
+                    if if_exists_flag:
+                        if_exists = 'append'
+                    temp_row_number = batch.to_sql(name=table_name, con=self.engine, schema=schema, if_exists=if_exists, index=index, index_label=index_label, dtype=self.dtypes, chunksize=chunksize, method=method)
+                    row_number += temp_row_number
+                    if_exists_flag = True
+            else:
+                print('under 1000 records to_sql does not support progress bar')
+                self.df = df
+                row_number = df.to_sql(name=table_name, con=self.engine, schema=schema, if_exists=if_exists, index=index, index_label=index_label, dtype=self.dtypes, chunksize=chunksize, method=method)
+        else:
+            self.df = df
+            row_number = df.to_sql(name=table_name, con=self.engine, schema=schema, if_exists=if_exists, index=index, index_label=index_label, dtype=self.dtypes, chunksize=chunksize, method=method)
 
         if primary_key != None:
             self.set_primary_key(table_name=table_name, schema=schema, column_name=primary_key)
@@ -393,6 +470,46 @@ class PySQL():
         """
         return pd.read_sql_query(query, con=self.engine, index_col=index_col, coerce_float=coerce_float, params=params, parse_dates=parse_dates, chunksize=chunksize, dtype=dtype)
     
+    @_log_decorator
+    def update_table(self, table_name, schema='dbo', update_key=None, update_value=None, print_query=False):
+        """
+            function to update table values with a [list of] dictionary.
+
+            Parameters
+            ----------
+            table_name : str
+                Name of SQL table in database.
+            schema : str, default dbo
+                Name of SQL schema in database to query (if database flavor
+                supports this). Uses default schema if dbo (default).
+            update_key : Dict or list of Dicts
+                json data in {'column_name': replace_key}
+            update_value : Dict or list of Dicts
+                json data in {'column1_name': to_replace_value, 'column2_name': to_replace_value}
+            
+            Returns
+            -------
+            1
+        """
+        if type(update_key)==list:
+            for update_value, update_key in zip(update_value, update_key):
+                QUERY = f'UPDATE {schema}.{table_name}'
+                QUERY += ' SET ' + ' , '.join([f" {i} = '{update_value[i]}' " for i in update_value.keys()])
+                QUERY += ' WHERE ' + ' and '.join([f" {i} = '{update_key[i]}' " for i in update_key.keys()])
+                with self.engine.connect() as conn:
+                    conn.execute(sql_text(QUERY))
+                    conn.commit()
+        else:    
+            QUERY = f'UPDATE {schema}.{table_name}'
+            QUERY += ' SET ' + ' , '.join([f" {i} = '{update_value[i]}' " for i in update_value.keys()])
+            QUERY += ' WHERE ' + ' and '.join([f" {i} = '{update_key[i]}' " for i in update_key.keys()])
+            if print_query:
+                print(QUERY)
+            with self.engine.connect() as conn:
+                conn.execute(sql_text(QUERY))
+                conn.commit()
+        return 1
+
     def date_has_time(self, column_data):
         non_null_values = column_data.dropna()
         for date in non_null_values:
@@ -413,6 +530,9 @@ class PySQL():
     
     def cutter_n_finder(self, dtype):
         n = str(dtype).split('(')[1][:-1]
+        if int(n) > 3800:
+            print('cutted ', str(n))
+            return 3800
         return int(n)
             
     def text_cutter(self, df):
@@ -442,11 +562,15 @@ class PySQL():
             
         primary_key_dtype = str(self.dtypes[column_name])
         QUERY = f"""ALTER TABLE {table_name} alter column {column_name} {primary_key_dtype} NOT NULL"""
-        self.engine.execute(QUERY)   
+        with self.engine.connect() as conn:
+            conn.execute(sql_text(QUERY))
+            conn.commit()
         QUERY = f"""ALTER TABLE {table_name}
                     ADD PRIMARY KEY ({column_name});"""
-        self.engine.execute(QUERY)
-        print('primary key sets on {column_name} with out any error')
+        with self.engine.connect() as conn:
+            conn.execute(sql_text(QUERY))
+            conn.commit()
+        print(f'primary key sets on {column_name} with out any error')
         
     @_log_decorator 
     def create_dtypes(self, dtype_dict, table_name, schema=None):
@@ -485,8 +609,10 @@ class PySQL():
             self.dtypes = self.dtypes.to_dict('records')[0]['dtypes_str']
             self.dtypes = eval(self.dtypes)
             self.dtypes = {i:eval(self.dtypes[i].replace('BOOLEAN', 'Boolean')) for i in self.dtypes} 
+            return True
         except:
             self.Error = "can't load dtypes table from database please try run PySQL.create_dtypes() first.  using Table_analyzer is suggested :) "
             raise Exception(self.Error)
+            return False
        
         
